#pragma once

#include <TNL/Meshes/MeshBuilder.h>
#include <TNL/Meshes/Mesh.h>
#include <TNL/Meshes/DefaultConfig.h>
#include <TNL/Meshes/Topologies/Tetrahedron.h>

template< typename Cell,
          int SpaceDimension = Cell::dimension,
          typename Real = double,
          typename GlobalIndex = int,
          typename LocalIndex = GlobalIndex >
struct MinimalConfig
: public TNL::Meshes::DefaultConfig< Cell, SpaceDimension, Real, GlobalIndex, LocalIndex >
{
	static constexpr bool subentityStorage( int entityDimension, int subentityDimension )
	{
		return subentityDimension == 0 && entityDimension == Cell::dimension;
	}

	static constexpr bool superentityStorage( int entityDimension, int superentityDimension )
	{
		return false;
	}

	static constexpr bool entityTagsStorage( int entityDimension )
	{
		return false;
	}

	static constexpr bool dualGraphStorage()
	{
		return false;
	}
};

template< typename Real, int SpaceDimension >
using PolygonalMesh = TNL::Meshes::Mesh< MinimalConfig<
                      TNL::Meshes::Topologies::Polygon,
                      SpaceDimension,
                      Real,
                      std::size_t,
                      short int > >;

template< typename Real >
using TetrahedralMesh = TNL::Meshes::Mesh< MinimalConfig<
                        TNL::Meshes::Topologies::Tetrahedron,
                        3,
                        Real,
                        std::size_t,
                        short int > >;

// overload for 2D or 3D surface meshes - polygonal meshes
template< typename SurfaceMesh >
PolygonalMesh< typename SurfaceMesh::Point::FT, 3 >  // FIXME: SurfaceMesh::Point::dimension() is not constexpr
cgal_mesh_to_tnl( const SurfaceMesh& surface )
{
	using MeshType = PolygonalMesh< typename SurfaceMesh::Point::FT, 3 >;  // FIXME: SurfaceMesh::Point::dimension() is not constexpr
	using PointType = typename MeshType::PointType;

	// initialize TNL mesh builder
	TNL::Meshes::MeshBuilder< MeshType > builder;
	builder.setEntitiesCount(surface.number_of_vertices(), surface.number_of_faces());

	// set points
	for( auto v : surface.vertices() ) {
		PointType point;
		for( int i = 0; i < PointType::getSize(); i++ )
			point[i] = surface.point(v)[i];
		builder.setPoint(v, point);
	}

	// count vertices per cell
	typename MeshType::MeshTraitsType::NeighborCountsArray cornersCounts(surface.number_of_faces());
	for( auto f : surface.faces() ) {
		const auto halfedge = surface.halfedge(f);
		typename MeshType::LocalIndexType count = 0;
		for( auto v : vertices_around_face(halfedge, surface) )
			count++;
		cornersCounts[f] = count;
	}
	builder.setCellCornersCounts(cornersCounts);

	// set cell seeds
	for( auto f : surface.faces() ) {
		auto seed = builder.getCellSeed(f);
		const auto halfedge = surface.halfedge(f);
		typename MeshType::LocalIndexType index = 0;
		for( auto v : vertices_around_face(halfedge, surface) )
			seed.setCornerId(index++, v);
	}

	// build the mesh
	MeshType mesh;
	if (!builder.build(mesh))
		throw std::runtime_error("TNL's MeshBuilder failed");
	return mesh;
}

// overload for 3D triangulations - tetrahedral meshes
template< typename C3T3 >
TetrahedralMesh< typename C3T3::Triangulation::Geom_traits::FT >
cgal_mesh_to_tnl( const C3T3& c3t3 )
{
	using MeshType = TetrahedralMesh< typename C3T3::Triangulation::Geom_traits::FT >;
	using PointType = typename MeshType::PointType;

	// NOTE: this function cannot work with the underlying triangulation alone:
	// - c3t3's triangulation contains even points outside the domain (on the
	//   bounding ball).
	// - Moreover, the triangulation may be non-regular after the remeshing
	//   algorithm, i.e. the empty-sphere property may be violated (there may
	//   additional points in the circumsphere of a cell). Especially on highly
	//   non-convex domains, the point may be actually inside the domain, so it
	//   is not counted in c3t3.number_of_far_points() and its in_dimension()
	//   may be > -1.
	// - Consequently, the mesh builder may fail with the error: Some points
	//   were not used for cells.
	// Hence, we need to first find the points that are actually used for cells:
	std::unordered_map<typename C3T3::Triangulation::Vertex_handle, std::size_t> V;
	std::size_t idx = 0;
	for( auto cit = c3t3.cells_in_complex_begin();
		cit != c3t3.cells_in_complex_end();
		cit++ )
	{
		for (int i = 0; i < 4; i++)
			if (V.count(cit->vertex(i)) == 0)
				V[cit->vertex(i)] = idx++;
	}

	// initialize TNL mesh builder
	TNL::Meshes::MeshBuilder< MeshType > builder;
	builder.setEntitiesCount(idx, c3t3.number_of_cells());

	// set points
	for( auto vit : c3t3.triangulation().finite_vertex_handles() )
	{
		if (V.count(vit) == 0)
			continue;
		builder.setPoint(V[vit], {vit->point()[0], vit->point()[1], vit->point()[2]});
	}

	// set cell seeds
	idx = 0;
	for( auto cit = c3t3.cells_in_complex_begin();
		cit != c3t3.cells_in_complex_end();
		cit++ )
	{
		auto seed = builder.getCellSeed(idx++);
		for (int i = 0; i < 4; i++)
			seed.setCornerId(i, V[cit->vertex(i)]);
	}

	// build the mesh
	MeshType mesh;
	if (!builder.build(mesh))
		throw std::runtime_error("TNL's MeshBuilder failed");
	return mesh;
}
