#include <filesystem>
#include <string>
#include <sstream>

#include <CGAL/Polygon_mesh_processing/IO/polygon_mesh_io.h>
#include <CGAL/Polygon_mesh_processing/triangulate_faces.h>
#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Mesh_triangulation_3.h>
#include <CGAL/Mesh_complex_3_in_triangulation_3.h>
#include <CGAL/Mesh_criteria_3.h>
#include <CGAL/Polyhedron_3.h>
#include <CGAL/Polyhedral_mesh_domain_with_features_3.h>
#include <CGAL/make_mesh_3.h>
#include <CGAL/refine_mesh_3.h>
#include <CGAL/tetrahedral_remeshing.h>

#include <TNL/Meshes/Writers/VTUWriter.h>
#include "cgal-mesh-to-tnl.h"

#include "lyra/lyra.hpp"

template <typename T>
std::string fmt_default(T value)
{
	std::stringstream str;
	str << " Default value: " << value << ".";
	return str.str();
}

struct Parameters
{
// pipeline options
	bool detect_features = false;
	double features_angle_bound = 60;

// meshing parameters
// - https://doc.cgal.org/latest/Mesh_3/index.html#Mesh_3TheMeshingCriteria
// - https://doc.cgal.org/latest/Mesh_3/classCGAL_1_1Mesh__criteria__3.html
	// edge_size: a scalar field (resp. a constant) providing a space varying
	// (resp. a uniform) upper bound for the lengths of curve edges. This
	// parameter has to be set to a positive value when 1-dimensional features
	// protection is used.
	double edge_size = 0;
	// facet_angle: a lower bound for the angles (in degrees) of the surface
	// mesh facets.
	double facet_angle = 25;
	// facet_size: a scalar field (resp. a constant) describing a space varying
	// (resp. a uniform) upper-bound or for the radii of the surface Delaunay
	// balls.
	double facet_size = 0;
	// facet_distance: a scalar field (resp. a constant) describing a space
	// varying (resp. a uniform) upper bound for the distance between the facet
	// circumcenter and the center of its surface Delaunay ball.
	double facet_distance = 0.125;
	// cell_radius_edge_ratio: an upper bound for the ratio between the
	// circumradius of a mesh tetrahedron and its shortest edge. There is a
	// theoretical bound for this parameter: the Delaunay refinement process is
	// guaranteed to terminate for values of cell_radius_edge_ratio bigger than 2.
	double cell_radius_edge_ratio = 3;
	// cell_size: a scalar field (resp. a constant) describing a space varying
	// (resp. a uniform) upper-bound for the circumradii of the mesh
	// tetrahedra.
	double cell_size = 0;

// optimizer options
// - https://doc.cgal.org/latest/Mesh_3/index.html#Mesh_3TheOptimizationParameters
	bool odt = false;
	bool lloyd = false;
	bool perturb = false;
	bool exude = false;
// optimizer parameters
// - https://doc.cgal.org/latest/Mesh_3/group__PkgMesh3Functions.html
	// odt_time_limit: a CPU time limit (in seconds) after which the smoother
	// is stopped. The default value is 0 and means that there is no time
	// limit.
	double odt_time_limit = 0;
	// odt_max_iteration_number: sets a limit on the number of performed
	// iterations. The default value of 0 means that there is no limit on the
	// number of performed iterations.
	std::size_t odt_max_iteration_number = 0;
	// odt_convergence: is a stopping criterion based on convergence: the
	// optimization process is stopped, when at the last iteration, the
	// displacement of any vertex is less than a given percentage of the length
	// of the shortest edge incident to that vertex. The parameter convergence
	// gives the threshold ratio.
	double odt_convergence = 0.02;
	// odt_freeze_bound: is designed to reduce running time of each
	// optimization iteration. Any vertex that has a displacement less than a
	// given percentage of the length (the of its shortest incident edge, is
	// frozen (i.e. is not relocated). The parameter freeze_bound gives the
	// threshold ratio.
	double odt_freeze_bound = 0.01;
	// odt_do_freeze: completes the freeze_bound parameter. If it is set to
	// true (default value), frozen vertices will not move anymore in next
	// iterations. Otherwise, at each iteration, any vertex that moves,
	// unfreezes all its incident vertices.
	bool odt_do_freeze = true;
	// lloyd_time_limit: a CPU time limit (in seconds) after which the smoother
	// is stopped. The default value is 0 and means that there is no time
	// limit.
	double lloyd_time_limit = 0;
	// lloyd_max_iteration_number: sets a limit on the number of performed
	// iterations. The default value of 0 means that there is no limit on the
	// number of performed iterations.
	std::size_t lloyd_max_iteration_number = 0;
	// lloyd_convergence: is a stopping criterion based on convergence: the
	// optimization process is stopped, when at the last iteration, the
	// displacement of any vertex is less than a given percentage of the length
	// of the shortest edge incident to that vertex. The parameter convergence
	// gives the threshold ratio.
	double lloyd_convergence = 0.02;
	// lloyd_freeze_bound: is designed to reduce running time of each
	// optimization iteration. Any vertex that has a displacement less than a
	// given percentage of the length (the of its shortest incident edge, is
	// frozen (i.e. is not relocated). The parameter freeze_bound gives the
	// threshold ratio.
	double lloyd_freeze_bound = 0.01;
	// lloyd_do_freeze: completes the freeze_bound parameter. If it is set to
	// true (default value), frozen vertices will not move anymore in next
	// iterations. Otherwise, at each iteration, any vertex that moves,
	// unfreezes all its incident vertices.
	bool lloyd_do_freeze = true;
	// perturb_time_limit: a CPU time limit (in seconds) after which the
	// perturber is stopped. The default value is 0 and means that there is no
	// time limit.
	double perturb_time_limit = 0;
	// perturb_sliver_bound: a targeted lower bound on dihedral angles (in
	// degrees) of mesh cells.
	double perturb_sliver_bound = 0;
	// exude_time_limit: a CPU time limit (in seconds) after which the exuder
	// is stopped. The default value is 0 and means that there is no time
	// limit.
	double exude_time_limit = 0;
	// exude_sliver_bound: a targeted lower bound on dihedral angles (in
	// degrees) of mesh cells.
	double exude_sliver_bound = 0;

// Multi-Material Isotropic Tetrahedral Remeshing
// - https://doc.cgal.org/latest/Tetrahedral_remeshing/index.html
// - https://doc.cgal.org/latest/Tetrahedral_remeshing/group__PkgTetrahedralRemeshingRef.html
	// remesh_iterations: The number of iterations for the remeshing algorithm.
	// Value of 0 disables remeshing.
	std::size_t remesh_iterations = 0;
	// remesh_target_edge_length: This parameter provides a uniform mesh
	// density target for the remeshing algorithm.
	double remesh_target_edge_length = 0;
	// remesh_boundaries: If false, none of the input volume boundaries can be
	// modified. Otherwise, the topology is preserved, but atomic operations
	// can be performed on the surfaces, and along feature polylines, such that
	// boundaries are remeshed.
	bool remesh_boundaries = true;

	void set_cli_options(lyra::cli & cli)
	{
		// pipeline options
		cli |= lyra::opt(detect_features)
				["-f"]["--detect-features"]
				("Before mesh generation, detect sharp features at polygonal "
				 "surface and insert them as features of the domain."
				 + fmt_default(detect_features));
		cli |= lyra::opt(features_angle_bound, "SCALAR")
				["--features-angle-bound"]
				("The maximum angle (in degrees) between the two normal vectors "
				 "of adjacent triangles. For an edge of the polyhedron, if the "
				 "angle between the two normal vectors of its incident facets is "
				 "bigger than the given bound, then the edge is considered as a "
				 "feature edge." + fmt_default(features_angle_bound));

		// meshing parameters
		cli |= lyra::opt(edge_size, "SCALAR")
				["--edge-size"]
				("A constant providing a uniform upper bound for the lengths of "
				 "curve edges. This parameter has to be set to a positive value when "
				 "--detect-features is used."
				 + fmt_default(edge_size));
		cli |= lyra::opt(facet_angle, "SCALAR")
				["--facet-angle"]
				("A lower bound for the angles (in degrees) of the surface mesh facets."
				 + fmt_default(facet_angle));
		cli |= lyra::opt(facet_size, "SCALAR")
				["--facet-size"]
				("A constant describing a uniform upper-bound or for the radii of "
				 "the surface Delaunay balls." + fmt_default(facet_size));
		cli |= lyra::opt(facet_distance, "SCALAR")
				["--facet-distance"]
				("A constant describing a uniform upper bound for the distance "
				 "between the facet circumcenter and the center of its surface "
				 "Delaunay ball." + fmt_default(facet_distance));
		cli |= lyra::opt(cell_radius_edge_ratio, "SCALAR")
				["--cell-radius-edge-ratio"]
				("An upper bound for the ratio between the circumradius of a mesh "
				 "tetrahedron and its shortest edge. There is a theoretical bound "
				 "for this parameter: the Delaunay refinement process is guaranteed "
				 "to terminate for values of cell_radius_edge_ratio bigger than 2."
				 + fmt_default(cell_radius_edge_ratio));
		cli |= lyra::opt(cell_size, "SCALAR")
				["--cell-size"]
				("A constant describing a uniform upper-bound for the circumradii "
				 "of the mesh tetrahedra." + fmt_default(cell_size));

		// optimizer options
		cli |= lyra::opt(odt)
				["-o"]["--odt"]
				("Activate the optimal Delaunay triangulation smoother - a mesh "
				 "optimizer based on the minimization of a global energy "
				 "function." + fmt_default(perturb));
		cli |= lyra::opt(lloyd)
				["-l"]["--lloyd"]
				("Activate the Lloyd smoother - a mesh optimizer based on the "
				 "minimization of a global energy function."
				 + fmt_default(perturb));
		cli |= lyra::opt(perturb)
				["-p"]["--perturb"]
				("Activate the perturber - a mesh optimizer that improves the "
				 "quality of a Delaunay mesh by changing the positions of some "
				 "vertices of the mesh." + fmt_default(perturb));
		cli |= lyra::opt(exude)
				["-e"]["--exude"]
				("Activate the exuder to exude slivers. The sliver exudation "
				 "process consists in optimizing the weights of vertices of the "
				 "weighted Delaunay triangulation in such a way that slivers "
				 "disappear and the quality of the mesh improves."
				 + fmt_default(exude));

		// optimizer parameters
		cli |= lyra::opt(odt_time_limit, "SCALAR")
				["--odt-time-limit"]
				("A CPU time limit (in seconds) after which the smoother is "
				 "stopped. The default value is 0 and means that there is no time "
				 "limit." + fmt_default(odt_time_limit));
		cli |= lyra::opt(odt_max_iteration_number, "N")
				["--odt-max-iteration-number"]
				("Sets a limit on the number of performed iterations. The default "
				 "value of 0 means that there is no limit on the number of performed "
				 "iterations." + fmt_default(odt_max_iteration_number));
		cli |= lyra::opt(odt_convergence, "SCALAR")
				["--odt-convergence"]
				("Is a stopping criterion based on convergence: the optimization "
				 "process is stopped, when at the last iteration, the displacement "
				 "of any vertex is less than a given percentage of the length of the "
				 "shortest edge incident to that vertex. The parameter convergence "
				 "gives the threshold ratio." + fmt_default(odt_convergence));
		cli |= lyra::opt(odt_freeze_bound, "SCALAR")
				["--odt-freeze-bound"]
				("Is designed to reduce running time of each optimization "
				 "iteration. Any vertex that has a displacement less than a given "
				 "percentage of the length (the of its shortest incident edge, is "
				 "frozen (i.e. is not relocated). The parameter freeze_bound gives "
				 "the threshold ratio." + fmt_default(odt_freeze_bound));
		cli |= lyra::opt(odt_do_freeze, "BOOL")
				["--odt-do-freeze"]
				("Completes the freeze_bound parameter. If it is set to true "
				 "(default value), frozen vertices will not move anymore in next "
				 "iterations. Otherwise, at each iteration, any vertex that moves, "
				 "unfreezes all its incident vertices." + fmt_default(odt_do_freeze));
		cli |= lyra::opt(lloyd_time_limit, "SCALAR")
				["--lloyd_time_limit"]
				("A CPU time limit (in seconds) after which the smoother is "
				 "stopped. The default value is 0 and means that there is no "
				 "time limit." + fmt_default(lloyd_time_limit));
		cli |= lyra::opt(lloyd_max_iteration_number, "N")
				["--lloyd-max-iteration-number"]
				("Sets a limit on the number of performed iterations. The default "
				 "value of 0 means that there is no limit on the number of performed "
				 "iterations." + fmt_default(lloyd_max_iteration_number));
		cli |= lyra::opt(lloyd_convergence, "SCALAR")
				["--lloyd-convergence"]
				("Is a stopping criterion based on convergence: the optimization "
				 "process is stopped, when at the last iteration, the displacement "
				 "of any vertex is less than a given percentage of the length of the "
				 "shortest edge incident to that vertex. The parameter convergence "
				 "gives the threshold ratio." + fmt_default(lloyd_convergence));
		cli |= lyra::opt(lloyd_freeze_bound, "SCALAR")
				["--lloyd-freeze-bound"]
				("Is designed to reduce running time of each optimization "
				 "iteration. Any vertex that has a displacement less than a given "
				 "percentage of the length (the of its shortest incident edge, is "
				 "frozen (i.e. is not relocated). The parameter freeze_bound gives "
				 "the threshold ratio." + fmt_default(lloyd_freeze_bound));
		cli |= lyra::opt(lloyd_do_freeze, "BOOL")
				["--lloyd-do-freeze"]
				("Completes the freeze_bound parameter. If it is set to true "
				 "(default value), frozen vertices will not move anymore in next "
				 "iterations. Otherwise, at each iteration, any vertex that moves, "
				 "unfreezes all its incident vertices." + fmt_default(lloyd_do_freeze));
		cli |= lyra::opt(perturb_time_limit, "SCALAR")
				["--perturb-time-limit"]
				("A CPU time limit (in seconds) after which the perturber is "
				 "stopped. The default value is 0 and means that there is no time "
				 "limit." + fmt_default(perturb_time_limit));
		cli |= lyra::opt(perturb_sliver_bound, "SCALAR")
				["--perturb-sliver-bound"]
				("A targeted lower bound on dihedral angles (in degrees) of mesh "
				 "cells." + fmt_default(perturb_sliver_bound));
		cli |= lyra::opt(exude_time_limit, "SCALAR")
				["--exude-time-limit"]
				("A CPU time limit (in seconds) after which the exuder is stopped. "
				 "The default value is 0 and means that there is no time limit."
				 + fmt_default(exude_time_limit));
		cli |= lyra::opt(exude_sliver_bound, "SCALAR")
				["--exude-sliver-bound"]
				("A targeted lower bound on dihedral angles (in degrees) of mesh "
				 "cells." + fmt_default(exude_sliver_bound));

		// remeshing parameters
		cli |= lyra::opt(remesh_iterations, "N")
				["--remesh-iterations"]
				("The number of iterations for the Multi-Material Isotropic "
				 "Tetrahedral Remeshing algorithm. Value of 0 disables "
				 "remeshing." + fmt_default(remesh_iterations));
		cli |= lyra::opt(remesh_target_edge_length, "SCALAR")
				["--remesh-target-edge-length"]
				("This parameter provides a uniform mesh density target for the "
				 "remeshing algorithm." + fmt_default(remesh_target_edge_length));
		cli |= lyra::opt(remesh_boundaries, "BOOL")
				["--remesh-boundaries"]
				("If false, none of the input volume boundaries can be "
				 "modified. Otherwise, the topology is preserved, but atomic "
				 "operations can be performed on the surfaces, and along feature "
				 "polylines, such that boundaries are remeshed."
				 + fmt_default(exude_time_limit));
	}

	void print_options(std::ostream & str)
	{
		str << "Pipeline parameters:\n";
		str << "- Detect features: " << detect_features << "\n";
		if (detect_features)
			str << "- Features angle bound: " << features_angle_bound << "\n";
		str << "Meshing parameters:\n";
		str << "- Edge size: " << edge_size << "\n";
		str << "- Facet angle: " << facet_angle << "\n";
		str << "- Facet size: " << facet_size << "\n";
		str << "- Facet distance: " << facet_distance << "\n";
		str << "- Cell radius-edge ratio: " << cell_radius_edge_ratio << "\n";
		str << "- Cell size: " << cell_size << "\n";
		str << "Optimizer parameters:\n";
		str << "- Run ODT smoother: " << odt << "\n";
		if (odt) {
			str << "  - Time limit: " << odt_time_limit << " seconds\n";
			str << "  - Maximum iterations: " << odt_max_iteration_number << "\n";
			str << "  - Convergence threshold: " << odt_convergence << "\n";
			str << "  - Freeze bound: " << odt_freeze_bound << "\n";
			str << "  - Activate vertex freezing: " << odt_do_freeze << "\n";
		}
		str << "- Run Lloyd smoother: " << lloyd << "\n";
		if (lloyd) {
			str << "  - Time limit: " << lloyd_time_limit << " seconds\n";
			str << "  - Maximum iterations: " << lloyd_max_iteration_number << "\n";
			str << "  - Convergence threshold: " << lloyd_convergence << "\n";
			str << "  - Freeze bound: " << lloyd_freeze_bound << "\n";
			str << "  - Activate vertex freezing: " << lloyd_do_freeze << "\n";
		}
		str << "- Run perturber: " << perturb << "\n";
		if (perturb) {
			str << "  - Time limit: " << perturb_time_limit << " seconds\n";
			str << "  - Sliver bound: " << perturb_sliver_bound << "°\n";
		}
		str << "- Run exuder: " << exude << "\n";
		if (exude) {
			str << "  - Time limit: " << exude_time_limit << " seconds\n";
			str << "  - Sliver bound: " << exude_sliver_bound << "°\n";
		}
		str << "- Number of remeshing iterations: " << remesh_iterations << "\n";
		if (remesh_iterations > 0) {
			str << "  - Remeshing target edge length: " << remesh_target_edge_length << "\n";
			str << "  - Remesh boundaries: " << remesh_boundaries << "\n";
		}
		str.flush();
	}
};

template< typename Polyhedron >
void readPolyhedron(Polyhedron & polyhedron, const std::filesystem::path & input_file, bool verbose = true)
{
	// the function from the Polygon_mesh_processing does some repairs and orientations
	// to obtain a valid polygon mesh
	// https://doc.cgal.org/latest/Polygon_mesh_processing/group__PMP__IO__grp.html
	if (!CGAL::Polygon_mesh_processing::IO::read_polygon_mesh(input_file, polyhedron, CGAL::parameters::verbose(verbose)))
		throw std::runtime_error("Could not read file " + input_file.string());

	if (!CGAL::is_triangle_mesh(polyhedron)) {
		if (verbose)
			std::cout << "Input surface mesh is not triangular, calling triangulate_faces..." << std::endl;
		CGAL::Polygon_mesh_processing::triangulate_faces(polyhedron);
	}
}

void polyDomain(const std::filesystem::path & input_file,
				const std::filesystem::path & output_file,
				const Parameters & args)
{
	// Domain
	using K = CGAL::Exact_predicates_inexact_constructions_kernel;
	using Polyhedron = CGAL::Mesh_polyhedron_3<K>::type;
	using Mesh_domain = CGAL::Polyhedral_mesh_domain_with_features_3<K>;

	#ifdef CGAL_CONCURRENT_MESH_3
	using ConcurrencyTag = CGAL::Parallel_tag;
	#else
	using ConcurrencyTag = CGAL::Sequential_tag;
	#endif

	// Triangulation
	using Tr = CGAL::Mesh_triangulation_3<Mesh_domain, CGAL::Default, ConcurrencyTag>::type;
	using C3T3 = CGAL::Mesh_complex_3_in_triangulation_3<Tr, Mesh_domain::Corner_index, Mesh_domain::Curve_index>;

	// To avoid verbose function and named parameters call
	using namespace CGAL::parameters;

	// Validate input arguments
	if (args.edge_size < 0) throw std::invalid_argument("edge_size cannot be negative");
	if (args.facet_size < 0) throw std::invalid_argument("facet_size cannot be negative");
	if (args.cell_size < 0) throw std::invalid_argument("cell_size cannot be negative");
	if (args.detect_features && args.edge_size == 0)
		throw std::invalid_argument("edge_size must be positive when detection of sharp features is enabled");
	if (args.remesh_iterations > 0 && args.remesh_target_edge_length <= 0)
		throw std::invalid_argument("remesh_iterations is positive, but remesh_target_edge_length is not positive");

	Polyhedron casePoly;
	readPolyhedron(casePoly, input_file);

	// Create domain
	Mesh_domain domain(casePoly);

	// Get sharp features
	if (args.detect_features)
		domain.detect_features(args.features_angle_bound);

	// Criteria
	using MeshCriteria = CGAL::Mesh_criteria_3<Tr>;
	MeshCriteria criteria(
			edge_size = args.edge_size,
			facet_angle = args.facet_angle,
			facet_size = args.facet_size,
			facet_distance = args.facet_distance,
			cell_radius_edge_ratio = args.cell_radius_edge_ratio,
			cell_size = args.cell_size
		);

	// Mesh generation
	std::cout << "Meshing..." << std::endl;
	C3T3 c3t3 = CGAL::make_mesh_3<C3T3>(domain, criteria, no_perturb(), no_exude());

	// Mesh optimization
	if (args.odt) {
		std::cout << "Running ODT smoother..." << std::endl;
		CGAL::odt_optimize_mesh_3(c3t3, domain,
				time_limit = args.odt_time_limit,
				max_iteration_number = args.odt_max_iteration_number,
				convergence = args.odt_convergence,
				freeze_bound = args.odt_freeze_bound,
				do_freeze = args.odt_do_freeze
			);
	}
	if (args.lloyd) {
		std::cout << "Running Lloyd smoother..." << std::endl;
		CGAL::lloyd_optimize_mesh_3(c3t3, domain,
				time_limit = args.lloyd_time_limit,
				max_iteration_number = args.lloyd_max_iteration_number,
				convergence = args.lloyd_convergence,
				freeze_bound = args.lloyd_freeze_bound,
				do_freeze = args.lloyd_do_freeze
			);
	}
	if (args.perturb) {
		std::cout << "Running perturber..." << std::endl;
		CGAL::perturb_mesh_3(c3t3, domain,
				time_limit = args.perturb_time_limit,
				sliver_bound = args.perturb_sliver_bound
			);
	}
	if (args.exude) {
		std::cout << "Running exuder..." << std::endl;
		CGAL::exude_mesh_3(c3t3,
				time_limit = args.exude_time_limit,
				sliver_bound = args.exude_sliver_bound
			);
	}

	// Remeshing
	if (args.remesh_iterations > 0) {
		std::cout << "Remeshing..." << std::endl;
		// NOTE: passing c3t3 directly to the remeshing function is an undocumented feature of CGAL
		CGAL::tetrahedral_isotropic_remeshing(c3t3,
			args.remesh_target_edge_length,
			CGAL::parameters::number_of_iterations(args.remesh_iterations)
			.remesh_boundaries(args.remesh_boundaries)
		);
	}

	// Output to VTU
	// (we use TNL instead of CGAL because the latter uses the "appended" data
	// format which is difficult to parse for libraries except VTK itself)
	using MeshType = TetrahedralMesh<typename C3T3::Triangulation::Geom_traits::FT>;
	const MeshType tnlMesh = cgal_mesh_to_tnl(c3t3);
	std::ofstream out(output_file);
	using MeshWriter = TNL::Meshes::Writers::VTUWriter<MeshType>;
	MeshWriter writer(out);
	writer.template writeEntities<3>(tnlMesh);
}

int main(int argc, char * argv[])
{
	Parameters args;
	std::filesystem::path input_file;
	std::filesystem::path output_path = ".";
	bool show_help = false;

	auto cli = lyra::cli();
	cli |= lyra::help(show_help).description("3D mesh generator based on the CGAL library");

	args.set_cli_options(cli);

	// positional arguments (should be generally defined after optional arguments)
	cli |= lyra::arg(input_file, "input_file")
			("Input file (supported formats: OFF, STL, OBJ, PLY, VTP, GOCAD).").required();
	cli |= lyra::arg(output_path, "output_path")
			("Output path. If it is an existing directory, the stem of "
			 "input_file (i.e., base file name without suffix) is appended to "
			 "output_path. Otherwise, output_path denotes the output file and "
			 "it must end with the \".vtu\" suffix." + fmt_default(output_path));

	auto result = cli.parse({ argc, argv });

	// check that the arguments where valid
	if (!result) {
		std::cerr << "Error in command line: " << result.message() << std::endl;
		std::cerr << cli << std::endl;
		return EXIT_FAILURE;
	}

	// show the help when asked for.
	if (show_help) {
		std::cout << cli << std::endl;
		return EXIT_SUCCESS;
	}

	// check if the input file exists
	if (!std::filesystem::exists(input_file)) {
		std::cerr << "Error: input_file \"" + input_file.string() + "\" does not exist." << std::endl;
		return EXIT_FAILURE;
	}

	// check if output_path is an existing directory
	if (std::filesystem::is_directory(output_path))
		output_path /= (input_file.stem().string() + ".vtu");
	// ensure that output_path ends with ".vtu"
	else if (output_path.extension() != ".vtu") {
		std::cerr << "Error: output_path \"" + output_path.string() + "\" does not end with \".vtu\"." << std::endl;
		return EXIT_FAILURE;
	}

	std::cout << "Input file: " << input_file << "\n";
	std::cout << "Output file: " << output_path << "\n";
	args.print_options(std::cout);

	try {
		polyDomain(input_file, output_path, args);
	}
	catch (std::logic_error & e) {
		std::cerr << "Logic error: " << e.what() << std::endl;
		return EXIT_FAILURE;
	}
	catch (std::runtime_error & e) {
		std::cerr << "Runtime error: " << e.what() << std::endl;
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
