#include <filesystem>

#include <CGAL/Simple_cartesian.h>
#include <CGAL/Surface_mesh.h>
#include <CGAL/Polygon_mesh_processing/IO/polygon_mesh_io.h>

#include <TNL/Meshes/Writers/VTUWriter.h>
#include "cgal-mesh-to-tnl.h"

#include "lyra/lyra.hpp"

template <typename T>
std::string fmt_default(T value)
{
	std::stringstream str;
	str << " Default value: " << value << ".";
	return str.str();
}

static const std::map<std::string, std::string> format_to_extension = {
	{"OFF", ".off"},
	{"STL", ".stl"},
	{"OBJ", ".obj"},
	{"PLY", ".ply"},
	{"GOCAD", ".ts"},
	{"VTP", ".vtp"},
	{"VTU", ".vtu"},
};

int main(int argc, char* argv[])
{
	bool show_help = false;
	bool verbose = false;
	std::string output_format = "VTU";
	std::filesystem::path input_file;
	std::filesystem::path output_path;

	auto cli = lyra::cli();
	cli |= lyra::help(show_help).description("Polygonal mesh converter based on the CGAL library");
	cli |= lyra::opt(verbose)["-v"]["--verbose"]("Instruct CGAL to produce verbose output.");
	cli |= lyra::opt(output_format, "FORMAT")["-f"]["--format"]("Output file format. Options: OFF, STL, OBJ, PLY, GOCAD, VTP, VTU.");
	cli |= lyra::arg(input_file, "input_file")("Input file (supported formats: OFF, STL, OBJ, PLY, GOCAD, VTP).").required();
	cli |= lyra::arg(output_path, "output_path")
			("Output path. If it is an existing directory, the stem of "
			 "input_file (i.e., base file name without suffix) is appended to "
			 "output_path. Otherwise, output_path denotes the output file and "
			 "it must end with the suffix matching the output file format."
			 + fmt_default(output_path));

	auto result = cli.parse({ argc, argv });

	// check that the arguments where valid
	if (!result) {
		std::cerr << "Error in command line: " << result.message() << std::endl;
		std::cerr << cli << std::endl;
		return EXIT_FAILURE;
	}

	// show the help when asked for.
	if (show_help) {
		std::cout << cli << std::endl;
		return EXIT_SUCCESS;
	}

	// check output file format
	if (format_to_extension.count(output_format) == 0) {
		std::cerr << "Error: unsupported output file format: \"" << output_format << "\"." << std::endl;
		return EXIT_FAILURE;
	}
	const std::string output_suffix = format_to_extension.at(output_format);

	// check if the input file exists
	if (!std::filesystem::exists(input_file)) {
		std::cerr << "Error: input_file \"" + input_file.string() + "\" does not exist." << std::endl;
		return EXIT_FAILURE;
	}

	// check if output_path is an existing directory
	if (std::filesystem::is_directory(output_path))
		output_path /= (input_file.stem().string() + output_suffix);
	// ensure that output_path ends with ".vtu"
	else if (output_path.extension() != output_suffix) {
		std::cerr << "Error: output_path \"" + output_path.string() + "\" does not end with the expected file format suffix \""
		          << output_suffix << "\"." << std::endl;
		return EXIT_FAILURE;
	}

	std::cout << "Input file: " << input_file << "\n";
	std::cout << "Output file: " << output_path << "\n";

	using Kernel = CGAL::Simple_cartesian<double>;
	using Point = Kernel::Point_3;
	using SurfaceMesh = CGAL::Surface_mesh<Point>;
	SurfaceMesh mesh;

	// the function from the Polygon_mesh_processing does some repairs and orientations
	// to obtain a valid polygon mesh
	// https://doc.cgal.org/latest/Polygon_mesh_processing/group__PMP__IO__grp.html
	if (!CGAL::Polygon_mesh_processing::IO::read_polygon_mesh(input_file, mesh, CGAL::parameters::verbose(verbose))) {
		std::cerr << "Invalid input file." << std::endl;
		return EXIT_FAILURE;
	}

	if (output_format == "VTU") {
		// Output to VTU
		// (we use TNL instead of CGAL because the latter uses the "appended" data
		// format which is difficult to parse for libraries except VTK itself)
		using MeshType = PolygonalMesh<double, 3>;
		const MeshType tnlMesh = cgal_mesh_to_tnl(mesh);
		std::ofstream out(output_path);
		using MeshWriter = TNL::Meshes::Writers::VTUWriter<MeshType>;
		MeshWriter writer(out);
		writer.template writeEntities<2>(tnlMesh);
	}
	else {
		// https://doc.cgal.org/latest/BGL/group__PkgBGLIOFct.html#gafa143949a33371dc6df8307be1ab8a66
		CGAL::IO::write_polygon_mesh(output_path, mesh, CGAL::parameters::stream_precision(17));

		// check if the output was created
		if (!std::filesystem::exists(output_path)) {
			std::cerr << "Error: CGAL failed to create the output file." << std::endl;
			return EXIT_FAILURE;
		}
	}

	return EXIT_SUCCESS;
}
