#pragma once

#include <TNL/Meshes/MeshBuilder.h>
#include <TNL/Meshes/Mesh.h>
#include <TNL/Meshes/DefaultConfig.h>
#include <TNL/Meshes/Topologies/Polygon.h>
#include <TNL/Meshes/Topologies/Polyhedron.h>

#include <cinolib/meshes/abstract_polyhedralmesh.h>
#include <cinolib/meshes/abstract_polygonmesh.h>

template< typename Cell,
          int SpaceDimension = Cell::dimension,
          typename Real = double,
          typename GlobalIndex = int,
          typename LocalIndex = GlobalIndex >
struct MinimalConfig
: public TNL::Meshes::DefaultConfig< Cell, SpaceDimension, Real, GlobalIndex, LocalIndex >
{
	static constexpr bool subentityStorage( int entityDimension, int subentityDimension )
	{
		constexpr int D = Cell::dimension;
		if( subentityDimension == 0 && entityDimension >= D - 1 )
			return true;
		if( D == 3 && subentityDimension == D - 1 && entityDimension == D )
			return true;
		return false;
	}

	static constexpr bool superentityStorage( int entityDimension, int superentityDimension )
	{
		return false;
	}

	static constexpr bool entityTagsStorage( int entityDimension )
	{
		return false;
	}

	static constexpr bool dualGraphStorage()
	{
		return false;
	}
};

template< typename Real >
using PolygonalMesh = TNL::Meshes::Mesh< MinimalConfig<
						TNL::Meshes::Topologies::Polygon,
						2,
						Real,
						std::size_t,
						short int > >;

template< typename Real >
using PolyhedralMesh = TNL::Meshes::Mesh< MinimalConfig<
						TNL::Meshes::Topologies::Polyhedron,
						3,
						Real,
						std::size_t,
						short int > >;

template< typename _M, typename _V, typename _E, typename _P >
PolygonalMesh< double >
cinolib_mesh_to_tnl( const cinolib::AbstractPolygonMesh<_M, _V, _E, _P>& m_poly )
{
	using MeshType = PolygonalMesh< double >;
	using PointType = typename MeshType::PointType;
	using NeighborCountsArray = typename TNL::Meshes::MeshBuilder< MeshType >::NeighborCountsArray;

	// NOTE:
	// - Cinolib's polygonal mesh may contain some points which are not used
	//   for cells (e.g. when the domain features are not preserved).
	// - Consequently, the mesh builder may fail with the error: Some points
	//   were not used for cells.
	// Hence, we need to first find the points that are actually used for cells:
	std::unordered_map<std::size_t, std::size_t> V;
	std::size_t idx = 0;
	for( const std::vector<uint> & f : m_poly.vector_polys() )
	{
		for( std::size_t i = 0; i < f.size(); i++)
			if (V.count(f[i]) == 0)
				V[f[i]] = idx++;
	}

	const std::size_t nv = idx;  // m_poly.vector_verts().size();
	const std::size_t nc = m_poly.vector_polys().size();

	// initialize TNL mesh builder
	TNL::Meshes::MeshBuilder< MeshType > builder;
	builder.setEntitiesCount(nv, nc);

	// set points
	idx = 0;
	for( const cinolib::vec3d & v : m_poly.vector_verts() )
	{
		if (V.count(idx) == 0)
		{
			idx++;
			continue;
		}
		builder.setPoint(V[idx], {v.x(), v.y(), v.z()});
		idx++;
	}

	// set cell corners counts
	NeighborCountsArray cellCorners( nc );
	idx = 0;
	for( const std::vector<uint> & c : m_poly.vector_polys() )
		cellCorners[idx++] = c.size();
	builder.setCellCornersCounts(cellCorners);
	cellCorners.reset();

	// set cell seeds
	idx = 0;
	for( const std::vector<uint> & c : m_poly.vector_polys() )
	{
		auto seed = builder.getCellSeed(idx++);
		for( std::size_t i = 0; i < c.size(); i++ )
			seed.setCornerId(i, V[c[i]]);
	}

	// build the mesh
	MeshType mesh;
	if (!builder.build(mesh))
		throw std::runtime_error("TNL's MeshBuilder failed");
	return mesh;
}

template< typename _M, typename _V, typename _E, typename _F, typename _P >
PolyhedralMesh< double >
cinolib_mesh_to_tnl( const cinolib::AbstractPolyhedralMesh<_M, _V, _E, _F, _P>& m_poly )
{
	using MeshType = PolyhedralMesh< double >;
	using PointType = typename MeshType::PointType;
	using NeighborCountsArray = typename TNL::Meshes::MeshBuilder< MeshType >::NeighborCountsArray;

	// NOTE:
	// - Cinolib's polyhedral mesh may contain some points which are not used
	//   for cells (e.g. when the domain features are not preserved).
	// - Consequently, the mesh builder may fail with the error: Some points
	//   were not used for cells.
	// Hence, we need to first find the points that are actually used for cells:
	std::unordered_map<std::size_t, std::size_t> V;
	std::size_t idx = 0;
	for( const std::vector<uint> & f : m_poly.vector_faces() )
	{
		for( std::size_t i = 0; i < f.size(); i++)
			if (V.count(f[i]) == 0)
				V[f[i]] = idx++;
	}

	const std::size_t nv = idx;  // m_poly.vector_verts().size();
	const std::size_t nf = m_poly.vector_faces().size();
	const std::size_t nc = m_poly.vector_polys().size();

	// initialize TNL mesh builder
	TNL::Meshes::MeshBuilder< MeshType > builder;
	builder.setEntitiesCount(nv, nc, nf);

	// set points
	idx = 0;
	for( const cinolib::vec3d & v : m_poly.vector_verts() )
	{
		if (V.count(idx) == 0)
		{
			idx++;
			continue;
		}
		builder.setPoint(V[idx], {v.x(), v.y(), v.z()});
		idx++;
	}

	// set face corners counts
	NeighborCountsArray faceCorners( nf );
	idx = 0;
	for( const std::vector<uint> & f : m_poly.vector_faces() )
		faceCorners[idx++] = f.size();
	builder.setFaceCornersCounts(faceCorners);
	faceCorners.reset();

	// set face seeds
	idx = 0;
	for( const std::vector<uint> & f : m_poly.vector_faces() )
	{
		auto seed = builder.getFaceSeed(idx++);
		for( std::size_t i = 0; i < f.size(); i++ )
			seed.setCornerId(i, V[f[i]]);
	}

	// set cell corners counts
	NeighborCountsArray cellCorners( nc );
	idx = 0;
	for( const std::vector<uint> & c : m_poly.vector_polys() )
		cellCorners[idx++] = c.size();
	builder.setCellCornersCounts(cellCorners);
	cellCorners.reset();

	// set cell seeds
	idx = 0;
	for( const std::vector<uint> & c : m_poly.vector_polys() )
	{
		auto seed = builder.getCellSeed(idx++);
		for( std::size_t i = 0; i < c.size(); i++ )
			seed.setCornerId(i, c[i]);
	}

	// build the mesh
	MeshType mesh;
	if (!builder.build(mesh))
		throw std::runtime_error("TNL's MeshBuilder failed");
	return mesh;
}
