#include <filesystem>
#include <string>
#include <sstream>

#include <cinolib/meshes/tetmesh.h>
#include <cinolib/meshes/polygonmesh.h>
#include <cinolib/meshes/polyhedralmesh.h>
#include <cinolib/dual_mesh.h>

#include <TNL/Meshes/Writers/VTUWriter.h>
#include "cinolib-mesh-to-tnl.h"

#include "lyra/lyra.hpp"

template <typename T>
std::string fmt_default(T value)
{
	std::stringstream str;
	str << " Default value: " << value << ".";
	return str.str();
}

struct Parameters
{
	// pipeline options
	bool detect_features = false;
	double features_angle_bound = 60;

	void set_cli_options(lyra::cli & cli)
	{
		// pipeline options
		cli |= lyra::opt(detect_features)
				["-f"]["--detect-features"]
				("Before dual mesh generation, detect sharp features on the primal "
				 "mesh surface and preserve them on the dual mesh."
				 + fmt_default(detect_features));
		cli |= lyra::opt(features_angle_bound, "SCALAR")
				["--features-angle-bound"]
				("The maximum angle (in degrees) between the two normal vectors "
				 "of adjacent triangles. For an edge of the mesh, if the angle "
				 "between the two normal vectors of its incident facets is bigger "
				 "than the given bound, then the edge is considered as a feature "
				 "edge." + fmt_default(features_angle_bound));
	}

	void print_options(std::ostream & str)
	{
		str << "Pipeline parameters:\n";
		str << "- Detect features: " << detect_features << "\n";
		if (detect_features)
			str << "- Features angle bound: " << features_angle_bound << "\n";
		str.flush();
	}
};

void generateDual(
		const std::filesystem::path & input_file,
		const std::filesystem::path & output_file,
		const Parameters & args)
{
	// try to load the input as tetrahedral mesh
	cinolib::Tetmesh<> m_tet;
	m_tet.load(input_file.string().c_str());

	if( m_tet.num_polys() > 0 )
	{
		// detect feature edges
		if (args.detect_features) {
			// convert degrees to radians
			const double angle_threshold = args.features_angle_bound * 3.1415926535897932384626433 / 180.;
			m_tet.edge_mark_sharp_creases(angle_threshold);
		}

		// make polyhedral (voronoi) mesh
		cinolib::Polyhedralmesh<> m_poly;
		cinolib::dual_mesh(m_tet, m_poly, args.detect_features);

		// Output to VTU using the writer from TNL
		using MeshType = PolyhedralMesh<double>;
		const MeshType tnlMesh = cinolib_mesh_to_tnl(m_poly);
		std::ofstream out(output_file);
		using MeshWriter = TNL::Meshes::Writers::VTUWriter<MeshType>;
		MeshWriter writer(out);
		writer.template writeEntities<3>(tnlMesh);
	}
	else
	{
		// try to load the input as polygonal mesh
		std::cout << "Input mesh does not contain any 3D cell, loading as 2D polygonal mesh." << std::endl;
		cinolib::Polygonmesh<> m_2d;
		m_2d.load(input_file.string().c_str());

		// detect feature edges
		if (args.detect_features) {
			// convert degrees to radians
			const double angle_threshold = args.features_angle_bound * 3.1415926535897932384626433 / 180.;
			m_tet.edge_mark_sharp_creases(angle_threshold);
		}

		// make polygonal (voronoi) mesh
		cinolib::Polygonmesh<> m_poly;
		cinolib::dual_mesh(m_2d, m_poly, args.detect_features);

		// Output to VTU using the writer from TNL
		using MeshType = PolygonalMesh<double>;
		const MeshType tnlMesh = cinolib_mesh_to_tnl(m_poly);
		std::ofstream out(output_file);
		using MeshWriter = TNL::Meshes::Writers::VTUWriter<MeshType>;
		MeshWriter writer(out);
		writer.template writeEntities<2>(tnlMesh);
	}
}

int main(int argc, char * argv[])
{
	Parameters args;
	std::filesystem::path input_file;
	std::filesystem::path output_path = ".";
	bool show_help = false;

	auto cli = lyra::cli();
	cli |= lyra::help(show_help).description("polygonal and polyhedral mesh generator based on the cinolib library");

	args.set_cli_options(cli);

	// positional arguments (should be generally defined after optional arguments)
	cli |= lyra::arg(input_file, "input_file")
			("Input file of 2D triangular (or general polygonal) mesh, or a "
			 "3D tetrahedral mesh. Supported formats: for 2D: OFF, OBJ, STL; "
			 "for 3D: VTK, VTU, TET, MESH.").required();
	cli |= lyra::arg(output_path, "output_path")
			("Output path. If it is an existing directory, the stem of "
			 "input_file (i.e., base file name without suffix) is appended to "
			 "output_path. Otherwise, output_path denotes the output file and "
			 "it must end with the \".vtu\" suffix." + fmt_default(output_path));

	auto result = cli.parse({ argc, argv });

	// check that the arguments where valid
	if (!result) {
		std::cerr << "Error in command line: " << result.message() << std::endl;
		std::cerr << cli << std::endl;
		return EXIT_FAILURE;
	}

	// show the help when asked for.
	if (show_help) {
		std::cout << cli << std::endl;
		return EXIT_SUCCESS;
	}

	// check if the input file exists
	if (!std::filesystem::exists(input_file)) {
		std::cerr << "Error: input_file \"" + input_file.string() + "\" does not exist." << std::endl;
		return EXIT_FAILURE;
	}

	// check if output_path is an existing directory
	if (std::filesystem::is_directory(output_path))
		output_path /= (input_file.stem().string() + ".vtu");
	// ensure that output_path ends with ".vtu"
	else if (output_path.extension() != ".vtu") {
		std::cerr << "Error: output_path \"" + output_path.string() + "\" does not end with \".vtu\"." << std::endl;
		return EXIT_FAILURE;
	}

	std::cout << "Input file: " << input_file << "\n";
	std::cout << "Output file: " << output_path << "\n";
	args.print_options(std::cout);

	try {
		generateDual(input_file, output_path, args);
	}
	catch (std::logic_error & e) {
		std::cerr << "Logic error: " << e.what() << std::endl;
		return EXIT_FAILURE;
	}
	catch (std::runtime_error & e) {
		std::cerr << "Runtime error: " << e.what() << std::endl;
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
